import fastgpx


class TestLatLong:

    # fastgpx.LatLong.__init__

    def test_init_with_two_floats_positionals(self):
        latlong = fastgpx.LatLong(12.34, 56.78)
        assert latlong.latitude == 12.34
        assert latlong.longitude == 56.78
        assert latlong.elevation == 0.0

    def test_init_with_three_floats_positionals(self):
        latlong = fastgpx.LatLong(12.345678, 56.789012, 89.2)
        assert latlong.latitude == 12.345678
        assert latlong.longitude == 56.789012
        assert latlong.elevation == 89.2

    def test_init_with_two_integers_positionals(self):
        latlong = fastgpx.LatLong(12, 56)
        assert latlong.latitude == 12.0
        assert latlong.longitude == 56.0
        assert latlong.elevation == 0.0

    def test_init_with_three_integers_positionals(self):
        latlong = fastgpx.LatLong(12, 56, 89)
        assert latlong.latitude == 12.0
        assert latlong.longitude == 56.0
        assert latlong.elevation == 89.0

    def test_init_with_two_floats_keywords(self):
        latlong = fastgpx.LatLong(latitude=12.34, longitude=56.78)
        assert latlong.latitude == 12.34
        assert latlong.longitude == 56.78
        assert latlong.elevation == 0.0

    def test_init_with_three_floats_keywords(self):
        latlong = fastgpx.LatLong(latitude=12.345678, longitude=56.789012, elevation=89.2)
        assert latlong.latitude == 12.345678
        assert latlong.longitude == 56.789012
        assert latlong.elevation == 89.2

    def test_init_with_two_integers_keywords(self):
        latlong = fastgpx.LatLong(latitude=12, longitude=56)
        assert latlong.latitude == 12.0
        assert latlong.longitude == 56.0
        assert latlong.elevation == 0.0

    def test_init_with_three_integers_keywords(self):
        latlong = fastgpx.LatLong(latitude=12, longitude=56, elevation=89)
        assert latlong.latitude == 12.0
        assert latlong.longitude == 56.0
        assert latlong.elevation == 89.0

    # fastgpx.LatLong.__eq__

    def test_equality(self):
        latlong1 = fastgpx.LatLong(12.34, 56.78, 90.1)
        latlong2 = fastgpx.LatLong(12.34, 56.78, 90.1)
        latlong3 = fastgpx.LatLong(12.34, 56.78, 91.2)

        assert latlong1 == latlong2
        assert latlong1 != latlong3

    # fastgpx.LatLong.__repr__

    def test_repr(self):
        latlong = fastgpx.LatLong(12.34, 56.78, 89.2)
        repr_str = repr(latlong)
        assert repr_str == "fastgpx.LatLong(latitude=12.34, longitude=56.78, elevation=89.2)"

    # fastgpx.LatLong.__str__

    def test_str(self):
        latlong = fastgpx.LatLong(12.34, 56.78, 89.2)
        str_str = str(latlong)
        assert str_str == "LatLong(12.34, 56.78, 89.2)"



import pytest

import fastgpx

METERS_TOL = 1e-4


class TestGeo:

    # fastgpx.geo.haversine

    def test_haversine(self):
        path = 'gpx/test/debug-segment.gpx'
        gpx = fastgpx.parse(path)
        points = gpx.tracks[0].segments[0].points
        assert len(points) == 2
        point1 = points[0]
        point2 = points[1]
        distance = fastgpx.geo.haversine(point1, point2)
        assert distance == pytest.approx(1.3839, abs=METERS_TOL)

    def test_haversine_distance(self):
        path = 'gpx/test/debug-segment.gpx'
        gpx = fastgpx.parse(path)
        points = gpx.tracks[0].segments[0].points
        assert len(points) == 2
        point1 = points[0]
        point2 = points[1]
        lat1 = point1.latitude
        lon1 = point1.longitude
        lat2 = point2.latitude
        lon2 = point2.longitude
        distance = fastgpx.geo.haversine_distance(lat1, lon1, lat2, lon2)
        assert distance == pytest.approx(1.3839, abs=METERS_TOL)



from datetime import datetime

import fastgpx


class TestTimeBounds:

    # fastgpx.TimeBounds.__init__

    def test_init_defaults(self):
        bounds = fastgpx.TimeBounds()
        assert bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time is None
        assert bounds.end_time is None

    def test_init_with_start_time(self):
        start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        bounds = fastgpx.TimeBounds(start_time=start_time, end_time=None)
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time == start_time
        assert bounds.end_time is None

    def test_init_with_end_time(self):
        end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds = fastgpx.TimeBounds(start_time=None, end_time=end_time)
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time is None
        assert bounds.end_time == end_time

    def test_init_with_start_and_end_time(self):
        start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds = fastgpx.TimeBounds(start_time=start_time, end_time=end_time)
        assert not bounds.is_empty()
        assert bounds.is_range()
        assert bounds.start_time == start_time
        assert bounds.end_time == end_time

    # fastgpx.TimeBounds.add

    def test_add_single_time_point(self):
        bounds = fastgpx.TimeBounds()
        time_point = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        bounds.add(time_point)
        assert not bounds.is_empty()
        assert bounds.is_range()
        assert bounds.start_time == time_point
        assert bounds.end_time == time_point

    def test_add_second_time_point(self):
        bounds = fastgpx.TimeBounds()
        time_point1 = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        time_point2 = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds.add(time_point1)
        bounds.add(time_point2)
        assert not bounds.is_empty()
        assert bounds.is_range()
        assert bounds.start_time == time_point1
        assert bounds.end_time == time_point2

    def test_add_time_bounds(self):
        bounds1 = fastgpx.TimeBounds(
            start_time=datetime.fromisoformat("2025-06-20 08:07:28+00:00"),
            end_time=datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        )
        bounds2 = fastgpx.TimeBounds(
            start_time=datetime.fromisoformat("2025-06-25 10:00:00+00:00"),
            end_time=datetime.fromisoformat("2025-06-30 15:00:00+00:00")
        )
        bounds1.add(bounds2)
        assert not bounds1.is_empty()
        assert bounds1.is_range()
        assert bounds1.start_time == bounds1.start_time
        assert bounds1.end_time == bounds2.end_time

    # fastgpx.TimeBounds.start_time

    def test_start_time_set_new_time(self):
        bounds = fastgpx.TimeBounds()
        new_start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        bounds.start_time = new_start_time
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time == new_start_time
        assert bounds.end_time is None

    def test_start_time_set_none(self):
        bounds = fastgpx.TimeBounds(
            start_time=datetime.fromisoformat("2025-06-20 08:07:28+00:00"),
            end_time=datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        )
        bounds.start_time = None
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time is None
        assert bounds.end_time == datetime.fromisoformat("2025-06-27 13:37:00+00:00")

    # fastgpx.TimeBounds.end_time

    def test_end_time_set_new_time(self):
        bounds = fastgpx.TimeBounds()
        new_end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds.end_time = new_end_time
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time is None
        assert bounds.end_time == new_end_time

    def test_end_time_set_none(self):
        bounds = fastgpx.TimeBounds(
            start_time=datetime.fromisoformat("2025-06-20 08:07:28+00:00"),
            end_time=datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        )
        bounds.end_time = None
        assert not bounds.is_empty()
        assert not bounds.is_range()
        assert bounds.start_time == datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        assert bounds.end_time is None

    # fastgpx.TimeBounds.__eq__

    def test_equality(self):
        start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds1 = fastgpx.TimeBounds(start_time=start_time, end_time=end_time)
        bounds2 = fastgpx.TimeBounds(start_time=start_time, end_time=end_time)
        bounds3 = fastgpx.TimeBounds(start_time=start_time, end_time=None)

        assert bounds1 == bounds2
        assert bounds1 != bounds3

    # fastgpx.TimeBounds.__repr__

    def test_repr(self):
        start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds = fastgpx.TimeBounds(start_time=start_time, end_time=end_time)
        repr_str = repr(bounds)
        expected_str = (
            "fastgpx.TimeBounds(start_time=datetime.datetime(2025, 6, 20, 8, 7, 28, tzinfo=datetime.timezone.utc), "
            "end_time=datetime.datetime(2025, 6, 27, 13, 37, tzinfo=datetime.timezone.utc))"
        )
        assert repr_str == expected_str

    # fastgpx.TimeBounds.__str__

    def test_str(self):
        start_time = datetime.fromisoformat("2025-06-20 08:07:28+00:00")
        end_time = datetime.fromisoformat("2025-06-27 13:37:00+00:00")
        bounds = fastgpx.TimeBounds(start_time=start_time, end_time=end_time)
        str_str = str(bounds)
        expected_str = "TimeBounds(2025-06-20T08:07:28Z to 2025-06-27T13:37:00Z)"
        assert str_str == expected_str



from collections.abc import Sequence
import enum
from typing import overload

import fastgpx


class Precision(enum.Enum):
    Five = 5

    Six = 6

@overload
def encode(locations: Sequence[fastgpx.LatLong], precision: Precision = Precision.Five) -> str: ...

@overload
def encode(locations: Sequence[fastgpx.LatLong], precision: int = 5) -> str: ...

@overload
def decode(encoded: str, precision: Precision = Precision.Five) -> list[fastgpx.LatLong]: ...

@overload
def decode(locations: str, precision: int = 5) -> list[fastgpx.LatLong]: ...



import math
import os
import time
import timeit
import xml.etree.ElementTree as ET
from typing import Callable, TypedDict

import gpxpy
from gpxpy.geo import EARTH_RADIUS
from gpxpy.geo import length_2d, length_3d, Location

from lxml import etree

from colorama import Fore, Back

import fastgpx

# https://github.com/jedie/gpxpy/commit/b371de31826cfecc049a57178d168b04a7f6d0d8
# https://github.com/jedie/gpxpy/blob/b371de31826cfecc049a57178d168b04a7f6d0d8/gpxpy/geo.py#L38
# https://github.com/jedie/gpxpy/blob/b371de31826cfecc049a57178d168b04a7f6d0d8/gpxpy/geo.py#L166
EARTH_RADIUS = 6378.137 * 1000


def to_rad(x: float) -> float:
    return x / 180.0 * math.pi


def haversine(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    d_lat = to_rad(lat2 - lat1)
    d_lon = to_rad(lon2 - lon1)
    lat1_rad = to_rad(lat1)
    lat2_rad = to_rad(lat2)

    a = math.sin(d_lat / 2) ** 2 + math.cos(lat1_rad) * \
        math.cos(lat2_rad) * math.sin(d_lon / 2) ** 2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return EARTH_RADIUS * c  # Distance in meters


def get_gpx_files(path: str):
    gpx_files = []
    files = os.listdir(path)
    for filename in files:
        if filename.endswith(".gpx"):
            gpx_path = os.path.join(path, filename)
            gpx_files.append(gpx_path)
    return gpx_files


PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
GPX_PATH = os.path.join(PROJECT_ROOT, 'gpx/2024 Great Roadtrip')

# gpxpy


def read_gpxpy():
    total_length = 0.0
    gpx_files = get_gpx_files(GPX_PATH)
    for gpx_filepath in gpx_files:
        with open(gpx_filepath, 'r', encoding='utf-8') as gpx_file:
            gpx = gpxpy.parse(gpx_file)
            total_length += gpx.length_2d()
            # total_length += gpx.length_3d()
    print('gpxpy', total_length, 'meters')
    return total_length

# xml.etree


def xml_etree_calculate_segment_length(trkseg) -> float:
    points = []
    ns = {'gpx': 'http://www.topografix.com/GPX/1/1'}
    for trkpt in trkseg.findall('.//gpx:trkpt', ns):
        lat = float(trkpt.attrib['lat'])
        lon = float(trkpt.attrib['lon'])
        points.append(Location(lat, lon))
    return length_2d(points)


def xml_etree_calculate_gpx_length(file_path: str) -> float:
    tree = ET.parse(file_path)
    root = tree.getroot()
    ns = {'gpx': 'http://www.topografix.com/GPX/1/1'}
    total_distance = 0.0
    for trk in root.findall('.//gpx:trk', ns):
        for trkseg in trk.findall('.//gpx:trkseg', ns):
            total_distance += xml_etree_calculate_segment_length(trkseg)
    return total_distance


def read_xml_etree():
    total_length = 0.0
    gpx_files = get_gpx_files(GPX_PATH)
    for gpx_filepath in gpx_files:
        length = xml_etree_calculate_gpx_length(gpx_filepath)
        total_length += length
    print('xml_etree', total_length, 'meters')
    return total_length


# lxml


def lxml_calculate_segment_length(trkseg) -> float:
    points = []
    for trkpt in trkseg.findall('.//{http://www.topografix.com/GPX/1/1}trkpt'):
        lat = float(trkpt.attrib['lat'])
        lon = float(trkpt.attrib['lon'])
        points.append(Location(lat, lon))
    return length_2d(points)


def lxml_calculate_gpx_length(file_path: str) -> float:
    tree = etree.parse(file_path, None)
    root = tree.getroot()

    total_distance = 0.0

    for trk in root.findall('.//{http://www.topografix.com/GPX/1/1}trk'):
        for trkseg in trk.findall('.//{http://www.topografix.com/GPX/1/1}trkseg'):
            total_distance += lxml_calculate_segment_length(trkseg)

    return total_distance


def read_lxml():
    total_length = 0.0
    gpx_files = get_gpx_files(GPX_PATH)
    for gpx_filepath in gpx_files:
        length = lxml_calculate_gpx_length(gpx_filepath)
        total_length += length
    print('lxml', total_length, 'meters')
    return total_length


def read_fastgpx():
    total_length = 0.0
    gpx_files = get_gpx_files(GPX_PATH)
    for gpx_filepath in gpx_files:
        fullpath = os.path.abspath(gpx_filepath)
        gpx = fastgpx.parse(fullpath)
        length = gpx.length_2d()
        total_length += length
    print('fastgpx', total_length, 'meters')
    return total_length

# Benchmarks:


class Benchmark(TypedDict):
    name: str
    function: Callable[[], float]


benchmarks: list[Benchmark] = [
    # {'name': 'gpxpy', 'function': read_gpxpy}, # Slow!
    {'name': 'xml_etree', 'function': read_xml_etree},
    {'name': 'lxml', 'function': read_lxml},
    {'name': 'fastgpx', 'function': read_fastgpx},
]

# print(f"Python script PID: {os.getpid()}")
# print("Press Enter to continue...")
# input()

print()
print('Computing expected distance with gpxpy...' + Fore.LIGHTMAGENTA_EX)
# gpxpy is _very_ slow, so omitting from doing multiple benchmark runs with it.
t = time.perf_counter()
expected = read_gpxpy()
d = time.perf_counter()
e = d - t
print(Fore.LIGHTYELLOW_EX + f'gpxpy: {e:.6f} seconds')
print()

iterations = 3
print(Fore.CYAN + f'Running {len(benchmarks)} benchmarks with {iterations} iterations...')
for benchmark in benchmarks:
    func = benchmark['function']
    print()
    # Benchmark timings:
    print(Fore.LIGHTYELLOW_EX + f"Running {benchmark['name']} ..." + Fore.LIGHTBLACK_EX)
    execution_time = timeit.timeit(func, number=iterations)
    average_time = execution_time / iterations
    print(Fore.LIGHTYELLOW_EX +
          f"{benchmark['name']}: {execution_time:.6f} seconds (Average: {average_time:.6f} seconds)" + Fore.YELLOW)
    # Print info about result accuracy:
    result = func()
    name = benchmark['name'].split(' ')[0]
    indent = ' ' * len(name)
    print(Fore.YELLOW +
          f"{indent} {expected} meters (expected)")
    deviation = result - expected
    percent = (deviation / expected) * 100
    print(Fore.LIGHTMAGENTA_EX + f'Distance deviation: {deviation} meters ({percent:.4f}%)')
    if deviation != 0.0:
        print(Back.LIGHTRED_EX + Fore.WHITE + '  FAIL  ' + Back.RESET + Fore.RESET)

print(Fore.RESET)



import gpxpy
import polyline
import pytest

import fastgpx


@pytest.fixture
def gpx_path():
    return "gpx/2024 TopCamp/Connected_20240518_094959_.gpx"


@pytest.fixture
def expected_gpx(gpx_path: str):
    with open(gpx_path, 'r', encoding='utf-8') as gpx_file:
        gpx = gpxpy.parse(gpx_file)
    return gpx


class TestPolyline:

    # fastgpx.polyline.encode

    def test_encode_p5_segment(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]

        expected = polyline.encode(points, precision=5)
        actual = fastgpx.polyline.encode(
            gpx.tracks[0].segments[0].points,
            precision=fastgpx.polyline.Precision.Five)
        assert actual == expected

    def test_encode_p6_segment(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]

        expected = polyline.encode(points, precision=6)
        actual = fastgpx.polyline.encode(
            gpx.tracks[0].segments[0].points,
            precision=fastgpx.polyline.Precision.Six)
        assert actual == expected

    def test_encode_p5_segment_int_overload(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]

        expected = polyline.encode(points, precision=5)
        actual = fastgpx.polyline.encode(
            gpx.tracks[0].segments[0].points, precision=5)
        assert actual == expected

    def test_encode_p6_segment_int_overload(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]

        expected = polyline.encode(points, precision=6)
        actual = fastgpx.polyline.encode(
            gpx.tracks[0].segments[0].points, precision=6)
        assert actual == expected

    def test_encode_segment_int_overload_invalid_arguments(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = gpx.tracks[0].segments[0].points

        with pytest.raises(ValueError):
            fastgpx.polyline.encode(points, precision=4)

        with pytest.raises(ValueError):
            fastgpx.polyline.encode(points, precision=7)

    # fastgpx.polyline.decode

    def test_decode_p5_segment(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]
        polyline6 = polyline.encode(points, precision=5)

        expected = polyline.decode(polyline6, precision=5)
        result = fastgpx.polyline.decode(polyline6,
                                         precision=fastgpx.polyline.Precision.Five)
        actual = [(point.latitude, point.longitude) for point in result]
        assert actual == expected

    def test_decode_p6_segment(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]
        polyline6 = polyline.encode(points, precision=6)

        expected = polyline.decode(polyline6, precision=6)
        result = fastgpx.polyline.decode(polyline6,
                                         precision=fastgpx.polyline.Precision.Six)
        actual = [(point.latitude, point.longitude) for point in result]
        assert actual == expected

    def test_decode_p5_segment_int_overload(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]
        polyline6 = polyline.encode(points, precision=5)

        expected = polyline.decode(polyline6, precision=5)
        result = fastgpx.polyline.decode(polyline6, precision=5)
        actual = [(point.latitude, point.longitude) for point in result]
        assert actual == expected

    def test_decode_p6_segment_int_overload(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]
        polyline6 = polyline.encode(points, precision=6)

        expected = polyline.decode(polyline6, precision=6)
        result = fastgpx.polyline.decode(polyline6, precision=6)
        actual = [(point.latitude, point.longitude) for point in result]
        assert actual == expected

    def test_decode_segment_int_overload_invalid_arguments(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        points = [(point.latitude, point.longitude)
                  for point in gpx.tracks[0].segments[0].points]
        polyline6 = polyline.encode(points, precision=6)

        with pytest.raises(ValueError):
            fastgpx.polyline.decode(polyline6, precision=4)

        with pytest.raises(ValueError):
            fastgpx.polyline.decode(polyline6, precision=7)



from collections.abc import Sequence
import datetime
from typing import overload

from . import geo as geo, polyline as polyline


class TimeBounds:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, start_time: datetime.datetime | datetime.date | datetime.time | None, end_time: datetime.datetime | datetime.date | datetime.time | None) -> None: ...

    @property
    def start_time(self) -> datetime.datetime | None: ...

    @start_time.setter
    def start_time(self, start_time: datetime.datetime | datetime.date | datetime.time | None) -> None: ...

    @property
    def end_time(self) -> datetime.datetime | None: ...

    @end_time.setter
    def end_time(self, end_time: datetime.datetime | datetime.date | datetime.time | None) -> None: ...

    def is_empty(self) -> bool: ...

    def is_range(self) -> bool: ...

    @overload
    def add(self, datetime: datetime.datetime | datetime.date | datetime.time) -> None: ...

    @overload
    def add(self, timebounds: TimeBounds) -> None: ...

    def __eq__(self, arg: object, /) -> bool: ...

    def __repr__(self) -> str: ...

    def __str__(self) -> str: ...

class LatLong:
    """Represent ``<trkpt>`` data in GPX files."""

    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, latitude: float, longitude: float, elevation: float = 0.0) -> None: ...

    @property
    def latitude(self) -> float:
        """The latitude of the point. Decimal degrees, WGS84 datum."""

    @latitude.setter
    def latitude(self, arg: float, /) -> None: ...

    @property
    def longitude(self) -> float:
        """The longitude of the point. Decimal degrees, WGS84 datum."""

    @longitude.setter
    def longitude(self, arg: float, /) -> None: ...

    @property
    def elevation(self) -> float:
        """The elevation of the point in meters."""

    @elevation.setter
    def elevation(self, arg: float, /) -> None: ...

    def __eq__(self, arg: object, /) -> bool: ...

    def __repr__(self) -> str: ...

    def __str__(self) -> str: ...

class Bounds:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, min: LatLong, max: LatLong) -> None: ...

    @overload
    def __init__(self, min: tuple[float, float], max: tuple[float, float]) -> None: ...

    @overload
    def __init__(self, min: tuple[float, float, float], max: tuple[float, float, float]) -> None: ...

    @property
    def min(self) -> LatLong | None: ...

    @min.setter
    def min(self, arg: LatLong, /) -> None: ...

    @property
    def max(self) -> LatLong | None: ...

    @max.setter
    def max(self, arg: LatLong, /) -> None: ...

    def is_empty(self) -> bool: ...

    @overload
    def add(self, location: LatLong) -> None: ...

    @overload
    def add(self, bounds: Bounds) -> None: ...

    def max_bounds(self, bounds: Bounds) -> Bounds: ...

    @property
    def min_latitude(self) -> float | None:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXBounds.min_latitude``.
           Prefer ``min.latitude`` instead. :func:`min`
        """

    @min_latitude.setter
    def min_latitude(self, arg: float, /) -> None: ...

    @property
    def min_longitude(self) -> float | None:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXBounds.min_longitude``.
           Prefer ``min.longitude`` instead. :func:`min`
        """

    @min_longitude.setter
    def min_longitude(self, arg: float, /) -> None: ...

    @property
    def max_latitude(self) -> float | None:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXBounds.max_latitude``.
           Prefer ``max.latitude`` instead. :func:`max`
        """

    @max_latitude.setter
    def max_latitude(self, arg: float, /) -> None: ...

    @property
    def max_longitude(self) -> float | None:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXBounds.max_longitude``.
           Prefer ``max.longitude`` instead. :func:`max`
        """

    @max_longitude.setter
    def max_longitude(self, arg: float, /) -> None: ...

    def __eq__(self, arg: object, /) -> bool: ...

    def __repr__(self) -> str: ...

    def __str__(self) -> str: ...

class Segment:
    """Represent ``<trkseg>`` data in GPX files."""

    def __init__(self) -> None: ...

    @property
    def points(self) -> list[LatLong]: ...

    @points.setter
    def points(self, arg: Sequence[LatLong], /) -> None: ...

    def bounds(self) -> Bounds: ...

    def get_bounds(self) -> Bounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXTrackSegment.get_bounds``.
           Prefer :func:`bounds` instead.
        """

    def time_bounds(self) -> TimeBounds: ...

    def get_time_bounds(self) -> TimeBounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXTrackSegment.get_time_bounds``.
           Prefer :func:`time_bounds` instead.
        """

    def length_2d(self) -> float:
        """Distance in meters."""

    def length_3d(self) -> float:
        """Distance in meters."""

    def __repr__(self) -> str: ...

class Track:
    """Represent ``<trk>`` data in GPX files."""

    def __init__(self) -> None: ...

    @property
    def name(self) -> str | None: ...

    @name.setter
    def name(self, arg: str, /) -> None: ...

    @property
    def comment(self) -> str | None: ...

    @comment.setter
    def comment(self, arg: str, /) -> None: ...

    @property
    def description(self) -> str | None: ...

    @description.setter
    def description(self, arg: str, /) -> None: ...

    @property
    def number(self) -> int | None: ...

    @number.setter
    def number(self, arg: int, /) -> None: ...

    @property
    def type(self) -> str | None: ...

    @type.setter
    def type(self, arg: str, /) -> None: ...

    @property
    def segments(self) -> list[Segment]: ...

    @segments.setter
    def segments(self, arg: Sequence[Segment], /) -> None: ...

    def bounds(self) -> Bounds: ...

    def get_bounds(self) -> Bounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXTrack.get_bounds``.
           Prefer :func:`bounds` instead.
        """

    def time_bounds(self) -> TimeBounds: ...

    def get_time_bounds(self) -> TimeBounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPXTrack.get_time_bounds``.
           Prefer :func:`time_bounds` instead.
        """

    def length_2d(self) -> float:
        """Distance in meters."""

    def length_3d(self) -> float:
        """Distance in meters."""

    def __repr__(self) -> str: ...

class Gpx:
    """Represent ``<gpx>`` data in GPX files."""

    def __init__(self) -> None: ...

    @property
    def tracks(self) -> list[Track]: ...

    @tracks.setter
    def tracks(self, arg: Sequence[Track], /) -> None: ...

    @property
    def name(self) -> str | None: ...

    @name.setter
    def name(self, arg: str, /) -> None: ...

    def bounds(self) -> Bounds: ...

    def get_bounds(self) -> Bounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPX.get_bounds``.
           Prefer :func:`bounds` instead.
        """

    def time_bounds(self) -> TimeBounds: ...

    def get_time_bounds(self) -> TimeBounds:
        """
        .. warning::

           Compatibility with ``gpxpy.GPX.get_time_bounds``.
           Prefer :func:`time_bounds` instead.
        """

    def length_2d(self) -> float:
        """Distance in meters."""

    def length_3d(self) -> float:
        """Distance in meters."""

    def __repr__(self) -> str: ...

def parse(path: str) -> Gpx: ...



import datetime

import gpxpy
import pytest

import fastgpx


@pytest.fixture
def gpx_path():
    return "gpx/2024 TopCamp/Connected_20240518_094959_.gpx"


@pytest.fixture
def expected_gpx(gpx_path: str):
    with open(gpx_path, 'r', encoding='utf-8') as gpx_file:
        gpx = gpxpy.parse(gpx_file)
    return gpx


METERS_TOL = 1e-4


class TestGpx:

    # fastgpx.Gpx.length_2d

    def test_length2d(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        distance = gpx.length_2d()
        assert distance == pytest.approx(382952.7193, abs=METERS_TOL)

    # fastgpx.Gpx.name

    def test_gpx_name_missing(self):
        path = 'gpx/test/debug-segment.gpx'
        gpx = fastgpx.parse(path)
        assert gpx.name is None

    def test_gpx_name_empty_string(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        assert gpx.name == ''

    def test_gpx_name(self):
        path = 'gpx/test/two-points.gpx'
        gpx = fastgpx.parse(path)
        assert gpx.name == 'Two Point Segment'

    # fastgpx.Gpx.bounds

    def test_bounds(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        bounds = gpx.bounds()
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == pytest.approx(61.410713)
        assert bounds.min.longitude == pytest.approx(10.427408)
        assert bounds.max is not None
        assert bounds.max.latitude == pytest.approx(63.441189)
        assert bounds.max.longitude == pytest.approx(13.142774)

    # fastgpx.Gpx.time_bounds

    def test_time_bounds(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        time_bounds = gpx.time_bounds()
        assert not time_bounds.is_empty()

        # 2024-05-18T07:50:00Z
        assert time_bounds.start_time is not None
        assert time_bounds.start_time == datetime.datetime(
            year=2024, month=5, day=18, hour=7, minute=50, second=0, tzinfo=datetime.timezone.utc)
        assert time_bounds.start_time.tzinfo is datetime.timezone.utc
        assert time_bounds.start_time.isoformat() == '2024-05-18T07:50:00+00:00'
        assert time_bounds.start_time.timestamp() == 1716018600

        # 2024-05-18T16:46:18Z
        assert time_bounds.end_time is not None
        assert time_bounds.end_time == datetime.datetime(
            year=2024, month=5, day=18, hour=16, minute=46, second=18, tzinfo=datetime.timezone.utc)
        assert time_bounds.end_time.tzinfo is datetime.timezone.utc
        assert time_bounds.end_time.isoformat() == '2024-05-18T16:46:18+00:00'
        assert time_bounds.end_time.timestamp() == 1716050778

    # fastgpx.Gpx.__repr__

    def test_repr(self):
        gpx = fastgpx.parse("gpx/test/two-points.gpx")
        repr_str = repr(gpx)
        assert repr_str == "<fastgpx.Gpx(tracks: 1, name: 'Two Point Segment')>"

    def test_repr_no_filename(self):
        gpx = fastgpx.parse("gpx/test/debug-segment.gpx")
        repr_str = repr(gpx)
        assert repr_str == "<fastgpx.Gpx(tracks: 1)>"


class TestTrack:

    # fastgpx.Track.length_2d

    def test_simple_track_length2d(self):
        path = 'gpx/test/debug-segment.gpx'
        gpx = fastgpx.parse(path)
        # Assigning intermediate values for easier debug inspection.
        tracks = gpx.tracks
        track = tracks[0]
        distance = track.length_2d()
        assert distance == pytest.approx(1.3839, abs=METERS_TOL)

    def test_length2d(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        track = gpx.tracks[0]
        distance = track.length_2d()
        assert distance == pytest.approx(382952.7193, abs=METERS_TOL)

    # fastgpx.Track.__repr__

    def test_repr(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        track = gpx.tracks[0]
        repr_str = repr(track)
        assert repr_str == "<fastgpx.Track(segments: 9)>"


class TestSegment:

    # fastgpx.Segment.length_2d

    def test_simple_segment_length2d(self):
        path = 'gpx/test/debug-segment.gpx'
        gpx = fastgpx.parse(path)
        # Assigning intermediate values for easier debug inspection.
        tracks = gpx.tracks
        track = tracks[0]
        segments = track.segments
        segment = segments[0]
        distance = segment.length_2d()
        assert distance == pytest.approx(1.3839, abs=METERS_TOL)

    def test_length2d(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        track = gpx.tracks[0]
        segment = track.segments[0]
        distance = segment.length_2d()
        assert distance == pytest.approx(17809.2701, abs=METERS_TOL)

    # fastgpx.Segment.__repr__

    def test_repr(self, gpx_path: str):
        gpx = fastgpx.parse(gpx_path)
        segment = gpx.tracks[0].segments[0]
        repr_str = repr(segment)
        assert repr_str == "<fastgpx.Segment(points: 1326)>"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import logging
import os
import re
import sys
from typing import TYPE_CHECKING

import sphinx
import tomllib

if TYPE_CHECKING:
    import sphinx.application


PROJECT_ROOT = os.path.abspath('../../')

sys.path.insert(0, os.path.join(PROJECT_ROOT, 'src'))

# Path: docs/source/_ext
# ext_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "_ext"))
# sys.path.insert(0, ext_path)

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# Read the version from pyproject.toml
with open(os.path.join(PROJECT_ROOT, 'pyproject.toml'), 'rb') as f:
    pyproject = tomllib.load(f)
    pyproject_name = pyproject['project']['name']
    pyproject_version = pyproject['project']['version']

assert pyproject['project']['name'] is not None
assert pyproject['project']['version'] is not None

project = pyproject['project']['name']
author = 'Thomas Thomassen'
copyright = f'2024-2025, {author}'
release = pyproject['project']['version']

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',  # For NumPy and Google style docstrings
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
]

templates_path = ['_templates']
# exclude_patterns = []

autosummary_generate = True

# autodoc_typehints = "description"
# autodoc_typehints = "both"
autodoc_typehints = "signature"
autodoc_typehints_description_target = "all"

# https://github.com/tox-dev/sphinx-autodoc-typehints?tab=readme-ov-file#options
typehints_fully_qualified = False
always_document_param_types = True
always_use_bars_union = True


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']


logger = logging.getLogger(__name__)

# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-signature
# Emitted when autodoc has formatted a signature for an object. The event handler can return a new
# tuple (signature, return_annotation) to change what Sphinx puts into the output.


def autodoc_process_signature(app, what, name, obj, options, signature, return_annotation):
    # https://github.com/wjakob/nanobind/discussions/707#discussioncomment-11708375
    logger.debug('')
    logger.debug('autodoc_process_signature')
    logger.debug(f"What: \033[35m{what}\033[0m, Name: \033[36m{name}\033[0m")
    logger.debug(f"Obj: {obj}")
    logger.debug(f"signature : \033[32m{signature}\033[0m")
    logger.debug(f"Doc:\n\033[33m{obj.__doc__}\033[0m")
    logger.debug('')
    # Fix signatures to use "fastgpx." instead of "fastgpx.fastgpx."
    # https://github.com/sphinx-doc/sphinx/issues/10351
    # Only the first `@overload` is given to this callback.
    if signature:
        signature = signature.replace("fastgpx.fastgpx.", "fastgpx.")
    if return_annotation:
        return_annotation = return_annotation.replace(
            "fastgpx.fastgpx.", "fastgpx.")
    return signature, return_annotation


def autodoc_before_process_signature(app, obj, bound_method):
    logger.debug('')
    logger.debug('autodoc_before_process_signature')
    logger.debug(f"Obj: {obj}")
    logger.debug(f"Doc:\n{obj.__doc__}")
    logger.debug('')


def autodoc_process_docstring(app, what, name, obj, options, lines):
    # https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring
    # Emitted when autodoc has read and processed a docstring. lines is a list
    # of strings – the lines of the processed docstring – that the event handler
    # can modify in place to change what Sphinx puts into the output.
    logger.debug('')
    logger.debug('autodoc_process_docstring')
    logger.debug(f"What: \033[35m{what}\033[0m, Name: \033[36m{name}\033[0m")
    logger.debug(f"Obj: {obj}")
    logger.debug(f"Lines: \n\033[33m{lines}\033[0m")
    logger.debug('')
    if what == 'property' and lines:
        logger.debug('\033[35mFixing property docstring...\033[0m')
        # nanobind + sphinx + autodoc seems to generate a docstring for
        # properties that includes the signature line. The PrettyPropertyDocumenter
        # also tries to parse this line to extract the return type.
        # Here we want to remove that signature line from the docstring to avoid
        # duplication.
        #
        # Example sig_line:
        #   (self) -> fastgpx.fastgpx.LatLong | None
        #
        # for line in lines:
        #     match = re.match(r'^\(.*\)\s*->\s*(.*)$', line)
        # Delete the first line if it matches the signature pattern
        match = re.match(r'^\(.*\)\s*->\s*(.*)$', lines[0])
        if match:
            lines.pop(0)


def monkey_patch_property_documenter(app: sphinx.application.Sphinx):
    from sphinx.ext.autodoc import PropertyDocumenter

    class MonkeyPatchPropertyDocumenter(PropertyDocumenter):
        # before PropertyDocumenter
        priority = PropertyDocumenter.priority + 1

        def add_directive_header(self, sig: str) -> None:
            super().add_directive_header(sig)
            sourcename = self.get_sourcename()

            logger.debug(
                f"\033[35mMonkeyPatchPropertyDocumenter adding directive header for {self.object}\033[0m")

            lines = self.get_doc()
            if not lines:
                return

            sig_line = lines[0][0]

            # Example sig_line:
            #   (self) -> fastgpx.fastgpx.LatLong | None
            #
            # We want to extract the return type:
            match = re.match(r'^\(.*\)\s*->\s*(.*)$', sig_line)
            if not match:
                return

            types = match.group(1)
            if types:
                types = types.replace("fastgpx.fastgpx.", "fastgpx.")

            objrepr = types
            self.add_line('   :type: ' + objrepr, sourcename)

    app.add_autodocumenter(MonkeyPatchPropertyDocumenter, override=True)


def monkey_patch_sphinx_for_nanobind():
    # Patch based on this discussion:
    # https://github.com/wjakob/nanobind/discussions/707#discussioncomment-13540168
    #
    # A more complete monkey patch for nanobind objects in Sphinx:
    # https://github.com/wjakob/nanobind/discussions/707#discussioncomment-15036945
    import inspect
    from sphinx.util import inspect as sphinx_inspect

    # Sphinx inspects all objects in the module and tries to resolve their type
    # (attribute, function, class, module, etc.) by using its own functions in
    # `sphinx.util.inspect`. These functions misidentify certain nanobind
    # objects. We monkey patch those functions here.
    def mpatch_ismethod(object):
        if hasattr(object, '__name__') and type(object).__name__ == "nb_method":
            return True
        return inspect.ismethod(object)

    sphinx_inspect_isclassmethod = sphinx_inspect.isclassmethod

    def mpatch_isclassmethod(object, cls=None, name=None):
        if hasattr(object, '__name__') and type(object).__name__ == "nb_method":
            return False
        return sphinx_inspect_isclassmethod(object, cls, name)

    sphinx_inspect_isfunction = sphinx_inspect.isfunction

    def mpatch_isfunction(object):
        if hasattr(object, '__name__') and type(object).__name__ == "nb_func":
            return True
        return sphinx_inspect_isfunction(object)

    sphinx_inspect.ismethod = mpatch_ismethod
    sphinx_inspect.isclassmethod = mpatch_isclassmethod
    sphinx_inspect.isfunction = mpatch_isfunction


def setup(app: sphinx.application.Sphinx):
    if app.verbosity >= 1:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    app.connect('autodoc-process-docstring', autodoc_process_docstring)
    app.connect('autodoc-process-signature', autodoc_process_signature)
    # app.connect('autodoc-before-process-signature', autodoc_before_process_signature)

    monkey_patch_property_documenter(app)
    monkey_patch_sphinx_for_nanobind()



from fastgpx import LatLong
from fastgpx import polyline, geo
from fastgpx.polyline import Precision

locations = [
    LatLong(64, 10),  # This works. Types and functions are accessible.
    LatLong(66, 11),
]

encoded1 = polyline.encode(locations, precision=polyline.Precision.Six)
decoded1 = polyline.decode(encoded1, precision=polyline.Precision.Six)
print("Encoded:", encoded1)
print("Decoded:", decoded1)
print()

encoded2 = polyline.encode(locations, precision=Precision.Six)
decoded2 = polyline.decode(encoded2, precision=Precision.Six)
print("Encoded:", encoded2)
print("Decoded:", decoded2)
print()

distance = geo.haversine(locations[0], locations[1])
print("Distance:", distance)
print()



import os

import polyline
from colorama import Fore, Back

from docs.source.conf import PROJECT_ROOT
import fastgpx


def get_gpx_files(path: str):
    gpx_files = []
    files = os.listdir(path)
    for filename in files:
        if filename.endswith(".gpx"):
            gpx_path = os.path.join(path, filename)
            gpx_files.append(gpx_path)
    return gpx_files


PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
GPX_PATH = os.path.join(PROJECT_ROOT, 'gpx/2024 Great Roadtrip')


def benchmark_fastgpx():
    polylines = []
    for gpx_filepath in gpx_files:
        gpx = fastgpx.parse(gpx_filepath)
        for track in gpx.tracks:
            for segment in track.segments:
                points = segment.points
                encoded = fastgpx.polyline.encode(points, precision=6)
                polylines.append(encoded)
    return polylines


def benchmark_polyline():
    polylines = []
    for gpx_filepath in gpx_files:
        gpx = fastgpx.parse(gpx_filepath)
        for track in gpx.tracks:
            for segment in track.segments:
                # Note: Extra overhead because this type conversion is needed.
                points = [(point.latitude, point.longitude) for point in segment.points]
                encoded = polyline.encode(points, precision=6)
                polylines.append(encoded)
    return polylines


if __name__ == '__main__':
    import timeit

    iterations = 10

    gpx_files = get_gpx_files(GPX_PATH)
    print(Fore.LIGHTBLUE_EX + f'GPX path: {GPX_PATH}')
    print(Fore.LIGHTBLUE_EX + f'GPX files: {len(gpx_files)}')
    print(Fore.LIGHTBLUE_EX + f'Iterations: {iterations}')
    # print('\n'.join(gpx_files))

    print()
    print(Fore.LIGHTYELLOW_EX + 'benchmarking fastgpx.polyline.encode' + Fore.RESET)
    print(timeit.timeit("benchmark_fastgpx()", globals=locals(), number=iterations))

    print()
    print(Fore.LIGHTYELLOW_EX + 'benchmarking polyline.encode' + Fore.RESET)
    print(timeit.timeit("benchmark_polyline()", globals=locals(), number=iterations))

    print(Fore.RESET)



import fastgpx


def haversine(latlong1: fastgpx.LatLong, latlong2: fastgpx.LatLong) -> float:
    """Haversine distance returned in meters using ``osmium`` logic."""

def haversine_distance(latitude_1: float, longitude_1: float, latitude_2: float, longitude_2: float) -> float:
    """
    .. warning::

       Signature compatibility with ``gpxpy.geo.haversine_distance``.
       Prefer :func:`haversine` instead.

    .. important::

       While the signature matches ``gpxpy``, the implementation uses ``osmium`` logic    which may lead to slightly different results.
    """



import fastgpx


class TestBounds:

    # fastgpx.Bounds.__init__

    def test_init_defaults(self):
        bounds = fastgpx.Bounds()
        assert bounds.is_empty()
        assert bounds.min is None
        assert bounds.min is None
        assert bounds.max is None
        assert bounds.max is None

    def test_init_with_latlongs_positionals(self):
        ll1 = fastgpx.LatLong(-10, -5)
        ll2 = fastgpx.LatLong(30, 25)
        bounds = fastgpx.Bounds(ll1, ll2)
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.max is not None

        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0

        assert bounds.min.latitude == bounds.min_latitude
        assert bounds.min.longitude == bounds.min_longitude
        assert bounds.max.latitude == bounds.max_latitude
        assert bounds.max.longitude == bounds.max_longitude

    def test_init_with_latlongs_keywords(self):
        ll1 = fastgpx.LatLong(-10, -5)
        ll2 = fastgpx.LatLong(30, 25)
        bounds = fastgpx.Bounds(min=ll1, max=ll2)
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.max is not None

        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0

    def test_init_with_two_tuples_positionals(self):
        bounds = fastgpx.Bounds((-10, -5), (30, 25))
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max is not None
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0

    def test_init_with_two_tuples_keywords(self):
        bounds = fastgpx.Bounds(min=(-10, -5), max=(30, 25))
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max is not None
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0

    def test_init_with_three_tuples_positionals(self):
        bounds = fastgpx.Bounds((-10, -5, 68), (30, 25, 150))
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.min.elevation == 68.0
        assert bounds.max is not None
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0
        assert bounds.max.elevation == 150.0

    def test_init_with_three_tuples_keywords(self):
        bounds = fastgpx.Bounds(min=(-10, -5, 68), max=(30, 25, 150))
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.min.elevation == 68.0
        assert bounds.max is not None
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0
        assert bounds.max.elevation == 150.0

    # fastgpx.Bounds.__eq__

    def test_equality(self):
        bounds1 = fastgpx.Bounds((-10, -5, 68), (30, 25, 150))
        bounds2 = fastgpx.Bounds(min=(-10, -5, 68), max=(30, 25, 150))
        bounds3 = fastgpx.Bounds((-10, -5), (30, 25))

        assert bounds1 == bounds2
        assert bounds1 != bounds3

    # fastgpx.Bounds.add

    def test_add_latlong(self):
        bounds = fastgpx.Bounds()
        assert bounds.is_empty()

        ll1 = fastgpx.LatLong(-10, -5)
        bounds.add(ll1)
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min == ll1
        assert bounds.max is not None
        assert bounds.max == ll1

        ll2 = fastgpx.LatLong(30, 25)
        bounds.add(ll2)
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max is not None
        assert bounds.max.latitude == 30.0
        assert bounds.max.longitude == 25.0

    # fastgpx.Bounds.max_bounds

    def test_max_bounds(self):
        bounds1 = fastgpx.Bounds((-10, -5), (30, 25))
        bounds2 = fastgpx.Bounds((5, 30), (40, 20))
        bounds = bounds1.max_bounds(bounds2)
        assert not bounds.is_empty()
        assert bounds.min is not None
        assert bounds.min.latitude == -10.0
        assert bounds.min.longitude == -5.0
        assert bounds.max is not None
        assert bounds.max.latitude == 40.0
        assert bounds.max.longitude == 30.0

    # fastgpx.Bounds.__repr__

    def test_repr(self):
        ll1 = fastgpx.LatLong(-10.3, -5.2, 68.5)
        ll2 = fastgpx.LatLong(30.3, 25.7, 150.2)
        bounds = fastgpx.Bounds(ll1, ll2)
        repr_str = repr(bounds)
        assert repr_str == "fastgpx.Bounds(min=(-10.3, -5.2, 68.5), max=(30.3, 25.7, 150.2))"

    # fastgpx.Bounds.__str__

    def test_str(self):
        ll1 = fastgpx.LatLong(-10.3, -5.2, 68.5)
        ll2 = fastgpx.LatLong(30.3, 25.7, 150.2)
        bounds = fastgpx.Bounds(ll1, ll2)
        str_str = str(bounds)
        assert str_str == "Bounds(min=(-10.3, -5.2, 68.5), max=(30.3, 25.7, 150.2))"



# This script updates the expected GPX data used in C++ tests.

import json
from pathlib import Path

import fastgpx


def get_gpx_files(dir_path: Path):
    return sorted(dir_path.glob("*.gpx"))


GPX_ROOTS = [
    Path("gpx/2024 Great Roadtrip"),
    Path("gpx/2024 Sommertur"),
    Path("gpx/2024 TopCamp"),
]


def main():
    files = []
    for root in GPX_ROOTS:
        files.extend(get_gpx_files(root))

    data = []
    for gpx_path in files:
        print(gpx_path.as_posix())

        assert gpx_path.exists()

        gpx = fastgpx.parse(str(gpx_path.resolve()))

        gpx_data = {
            "path": gpx_path.as_posix(),
            "length2d": round(gpx.length_2d(), 4),
            "length3d": round(gpx.length_3d(), 4),
            "tracks": [],
        }

        gpx_time = gpx.time_bounds()
        if gpx_time.start_time and gpx_time.end_time:
            gpx_data["time_bounds"] = {
                "start_time": int(gpx_time.start_time.timestamp()),
                "end_time": int(gpx_time.end_time.timestamp()),
            }

        for track in gpx.tracks:
            track_data = {
                "length2d": round(track.length_2d(), 4),
                "length3d": round(track.length_3d(), 4),
                "segments": [],
            }

            track_time = track.time_bounds()
            if track_time.start_time and track_time.end_time:
                track_data["time_bounds"] = {
                    "start_time": int(track_time.start_time.timestamp()),
                    "end_time": int(track_time.end_time.timestamp()),
                }

            for segment in track.segments:
                segment_data = {
                    "length2d": round(segment.length_2d(), 4),
                    "length3d": round(segment.length_3d(), 4),
                }

                segment_time = segment.time_bounds()
                if segment_time.start_time and segment_time.end_time:
                    segment_data["time_bounds"] = {  # type: ignore
                        "start_time": int(segment_time.start_time.timestamp()),
                        "end_time": int(segment_time.end_time.timestamp()),
                    }

                track_data["segments"].append(segment_data)

            gpx_data["tracks"].append(track_data)

        data.append(gpx_data)

    output_path = Path("cpp/expected_gpx_data.json")
    output_path.parent.mkdir(parents=True, exist_ok=True)

    with output_path.open("w", encoding="utf-8") as json_file:
        json.dump(data, json_file, indent=2)
        json_file.write("\n")


if __name__ == "__main__":
    main()
